use base64::{engine::general_purpose, Engine};
use dircpy::*;
use helmzoo_lib::git::{CheckoutOptions, CommitOptions, Git, GitOptions, PushOptions};
use helmzoo_lib::template;
use serde::{Deserialize, Serialize};

use crate::config::ChartExtended;

use super::Target;

#[derive(Serialize, Deserialize, PartialEq, Debug, Clone)]
pub(crate) struct GitMirror {
    pub(crate) git_dir: Option<String>,
    pub(crate) url: String,
    pub(crate) path: String,
    #[serde(default = "default_branch")]
    pub(crate) branch: String,
    #[serde(default = "default_commit")]
    pub(crate) commit: String,
    pub(crate) rebase_to: Option<String>,
    #[serde(default = "default_git_bin")]
    pub(crate) git_bin: String,
}

fn default_commit() -> String {
    "helmuled {{ name }}-{{ version }}".to_string()
}

fn default_branch() -> String {
    "main".to_string()
}

fn default_git_bin() -> String {
    "git".to_string()
}

impl Target for GitMirror {
    fn push(
        &self,
        workdir_path: String,
        chart_path: String,
        chart_local: ChartExtended,
        dry_run: bool,
    ) -> Result<(), Box<dyn std::error::Error>> {
        let repo_path = match self.git_dir.clone() {
            Some(dir) => template::render(dir.clone(), &chart_local)?,
            None => general_purpose::STANDARD_NO_PAD.encode(self.url.clone()),
        };

        let git_instance = Git {
            url: template::render(self.url.clone(), &chart_local)?,
            repo_path: repo_path.clone(),
        };
        let git_opts = GitOptions::new(self.git_bin.clone(), Some(workdir_path.clone()));
        git_instance.clone(git_opts.clone())?;

        // Prepare branch
        let checkout_opts = CheckoutOptions {
            create: true,
            git_ref: template::render(self.branch.clone(), &chart_local)?,
        };

        git_instance.checkout(git_opts.clone(), checkout_opts)?;

        // Prepare path
        let path = template::render(self.path.clone(), &chart_local)?;
        let repo_local_full_path = format!("{}/{}/{}", workdir_path.clone(), repo_path, path);
        CopyBuilder::new(chart_path.clone(), repo_local_full_path.clone())
            .overwrite_if_size_differs(true)
            .run()?;

        // Prepare the commit message
        let commit_opts = CommitOptions {
            message: template::render(self.commit.clone(), &chart_local)?,
            add: true,
        };

        git_instance.commit(git_opts.clone(), commit_opts)?;

        if !dry_run {
            let force_push = match self.rebase_to {
                Some(_) => true,
                None => false,
            };
            let push_opts = PushOptions {
                rebase_to: self.rebase_to.clone(),
                force: force_push,
                brahcn: self.branch.clone(),
            };
            git_instance.push(git_opts.clone(), push_opts)?;
        }
        Ok(())
    }
}
